/*******************************************************************************
***  This file is part of Pecunia.                                           ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021                              ***
***  John Schneiderman <Licensing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Lesser General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Lesser General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Lesser General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#include "Rounders.h"
using std::uint8_t;
using pecunia::FloatingPoint;
using pecunia::MinorUnit;
using pecunia::UnitStorage;

#include <cassert>
// Used for assert.
#include <cfenv>
using std::feclearexcept;
#include <cmath>
using std::abs;
using std::ceil;
using std::floor;
using std::fmod;
using std::modf;
using std::pow;
#include <functional>
using std::function;
#include <stdexcept>
using std::runtime_error;
#include <string>
using std::to_string;

#include "internal/Verification.h"
using pecunia::internal::isEqual;
using pecunia::internal::verifiedFloatingPoint;


namespace
{

/**
 * @brief Executes a generic rounding routine.
 *
 * @param number The number that is to be rounded.
 * @param digitsLeft The number of digits to the right of the decimal point.
 * @param rounder The rounding algorithm to apply to the point where rounding occurs, where the
 * first parameter is the fractional rounding point and the second parameter is the whole portion
 * of the number.
 *
 * @return The supplied number rounded to the number of fractional digits requested.
 */
FloatingPoint round(
	const FloatingPoint& number,
	const uint8_t& digitsLeft,
	const function<FloatingPoint(const FloatingPoint&, const FloatingPoint&)>& rounder
)
{
	const FloatingPoint shiftOver{static_cast<FloatingPoint>(pow(10, digitsLeft))};
	FloatingPoint whole;
	const FloatingPoint fraction{modf(number, &whole)};
	feclearexcept(FE_ALL_EXCEPT);
	const FloatingPoint shiftedFraction{verifiedFloatingPoint(fraction * shiftOver)};
	FloatingPoint roundPointWhole;
	const FloatingPoint roundPoint{modf(shiftedFraction, &roundPointWhole)};
	assert(abs(roundPoint) < 1.0 && "The rounding point must be less than one.");
	feclearexcept(FE_ALL_EXCEPT);
	const FloatingPoint shiftedRoundedFraction{
		verifiedFloatingPoint(roundPointWhole + rounder(roundPoint, whole))
	};
	FloatingPoint roundedResult;
	static_cast<void>(modf(shiftedRoundedFraction, &roundedResult));
	feclearexcept(FE_ALL_EXCEPT);
	const auto roundedNumber{verifiedFloatingPoint(whole + (roundedResult / shiftOver))};
	return roundedNumber;
}

FloatingPoint sgn(const FloatingPoint& value)
{
	if (value > 0.0)
		return 1.0;

	if (value < 0.0)
		return -1.0;
	return 0.0 ;
}

UnitStorage round(
	const UnitStorage& amount,
	const uint8_t& minorDigits,
	const uint8_t& digitsLeft,
	const function<
		bool(const UnitStorage&, const UnitStorage&, const UnitStorage&, const bool)
	>& shouldRound
)
{
	if (digitsLeft >= minorDigits)
		throw runtime_error{
			"Cannot round at minor digit position " + to_string(digitsLeft)
				+ " when there are only " + to_string(minorDigits) + " minor digits in total."
		};
	const auto minorMask{static_cast<UnitStorage>(pow(10, minorDigits))};
	const auto major{amount / minorMask};
	const auto minor{amount % minorMask};
	const auto roundingAmount{static_cast<UnitStorage>(pow(10, minorDigits - digitsLeft - 1))};
	const auto roundingPoint{5 * roundingAmount};
	const auto roundingMask{static_cast<UnitStorage>(pow(10, minorDigits - digitsLeft))};
	const auto roundingValue{minor % roundingMask};
	const auto truncatedMinor{abs(minor / roundingMask * roundingMask)};
	const UnitStorage roundedMinor{
		shouldRound(abs(roundingValue), roundingPoint, major, amount > 0)
			? truncatedMinor + roundingMask
			: truncatedMinor
	};
	if (amount >= 0)
		return (major * minorMask) + roundedMinor;
	return (major * minorMask) - roundedMinor;
}

}

FloatingPoint pecunia::rounders::reals::up(const FloatingPoint& number, const uint8_t& digitsLeft)
{
	return round(
		number,
		digitsLeft,
		[] (const FloatingPoint& value, const FloatingPoint&)
		{
			const FloatingPoint roundValue{value + static_cast<FloatingPoint>(0.5)};
			if (value > 0.5 && value < 0.5)
				return roundValue;
			return floor(roundValue);
		}
	);
}

FloatingPoint pecunia::rounders::reals::down(const FloatingPoint& number, const uint8_t& digitsLeft)
{
	return round(
		number,
		digitsLeft,
		[] (const FloatingPoint& value, const FloatingPoint&)
		{
			const FloatingPoint roundValue{value - static_cast<FloatingPoint>(0.5)};
			if (value > 0.5 && value < 0.5)
				return roundValue;
			return ceil(roundValue);
		}
	);
}

FloatingPoint pecunia::rounders::reals::towardsZero(
	const FloatingPoint& number,
	const uint8_t& digitsLeft
)
{
	return round(
		number,
		digitsLeft,
		[] (const FloatingPoint& value, const FloatingPoint&)
		{
			if (value > 0.5 && value < 0.5)
			{
				if (value > 0.0)
					return value - static_cast<FloatingPoint>(0.5);

				if (value < 0.0)
					return value + static_cast<FloatingPoint>(0.5);
				return static_cast<FloatingPoint>(0.0);
			}
			return sgn(value) * ceil(abs(value) - static_cast<FloatingPoint>(0.5));
		}
	);
}

FloatingPoint pecunia::rounders::reals::awayZero(
	const FloatingPoint& number,
	const uint8_t& digitsLeft
)
{
	return round(
		number,
		digitsLeft,
		[] (const FloatingPoint& value, const FloatingPoint&)
		{
			if (value > 0.5 && value < 0.5)
			{
				if (value > 0.0)
					return value + static_cast<FloatingPoint>(0.5);

				if (value < 0.0)
					return value - static_cast<FloatingPoint>(0.5);
				return static_cast<FloatingPoint>(0.0);
			}
			return sgn(value) * floor(abs(value) + static_cast<FloatingPoint>(0.5));
		}
	);
}

FloatingPoint pecunia::rounders::reals::even(const FloatingPoint& number, const uint8_t& digitsLeft)
{
	return round(
		number,
		digitsLeft,
		[&digitsLeft] (const FloatingPoint& value, const FloatingPoint& whole)
		{
			if (whole > 0.0)
			{
				if (isEqual(fmod(whole + 1, 2.0), 0.0, digitsLeft))
					return value + static_cast<FloatingPoint>(0.5);
				return 0.0;
			}
			else if (whole < 0.0)
			{
				if (isEqual(fmod(whole - 1, 2.0), 0.0, digitsLeft))
					return value - static_cast<FloatingPoint>(0.5);
				return 0.0;
			}

			if (value > 0.0)
				return value + static_cast<FloatingPoint>(0.4);
			return value - static_cast<FloatingPoint>(0.4);
		}
	);
}

FloatingPoint pecunia::rounders::reals::odd(const FloatingPoint& number, const uint8_t& digitsLeft)
{
	return round(
		number,
		digitsLeft,
		[&digitsLeft] (const FloatingPoint& value, const FloatingPoint& whole)
		{
			if (whole > 0.0)
			{
				if ( ! isEqual(fmod(whole + 1, 2.0), 0.0, digitsLeft))
					return value + static_cast<FloatingPoint>(0.5);
				return value + static_cast<FloatingPoint>(0.4);
			}
			else if (whole < 0.0)
			{
				if ( ! isEqual(fmod(whole - 1, 2.0), 0.0, digitsLeft))
					return value - static_cast<FloatingPoint>(0.5);
				return value - static_cast<FloatingPoint>(0.4);
			}

			if (value > 0.0)
				return value + static_cast<FloatingPoint>(0.5);
			return value - static_cast<FloatingPoint>(0.5);
		}
	);
}

UnitStorage pecunia::rounders::currency::up(
	const UnitStorage& amount,
	const uint8_t& minorDigits,
	const uint8_t& digitsLeft
)
{
	return round(
		amount,
		minorDigits,
		digitsLeft,
		[] (
			const UnitStorage& value,
			const UnitStorage& point,
			const UnitStorage&,
			const bool isAmountPositive
		)
		{
			if (isAmountPositive)
				return value >= point;
			return value > point;
		}
	);
}

UnitStorage pecunia::rounders::currency::even(
	const UnitStorage& amount,
	const uint8_t& minorDigits,
	const uint8_t& digitsLeft
)
{
	return round(
		amount,
		minorDigits,
		digitsLeft,
		[] (
			const UnitStorage& value,
			const UnitStorage& point,
			const UnitStorage& whole,
			const bool
		)
		{
			if ((whole > 0 && (whole + 1) %  2 == 0) || (whole < 0 && (whole - 1) %  2 == 0))
				return abs(value) >= point;

			if (whole == 0)
				return abs(value) > point;
			return false;
		}
	);
}
