/*******************************************************************************
***  This file is part of Pecunia.                                           ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021                              ***
***  John Schneiderman <Licensing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Lesser General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Lesser General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Lesser General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#include "Adjustments.h"
using std::function;
using std::int32_t;
using std::uint8_t;
using pecunia::precision;
using pecunia::MinorUnit;
using pecunia::FloatingPoint;
using pecunia::FloatingPointRounderFunction;
using pecunia::UnitStorage;

#include <cassert>
// using assert;
#include <cfenv>
using std::feclearexcept;
#include <cmath>
using std::abs;
#include <stdexcept>
using std::overflow_error;

#include "MoneySize.hpp"
using pecunia::internal::extraFloatingPointPrecision;
#include "pecunia/Codes.h"
using pecunia::currency::Iso4217Codes;
#include "pecunia/Money.h"
using pecunia::currency::Money;
using pecunia::currency::to_string;
#include "Verification.h"
using pecunia::internal::isEqual;
using pecunia::internal::verifiedFloatingPoint;


UnitStorage pecunia::currency::internal::fromFloatingPointMajor(
	const FloatingPoint& value,
	const Iso4217Codes& code,
	const uint8_t currencyPrecision,
	const FloatingPointRounderFunction& rounder
)
{
	assert(rounder != nullptr && "The floating point rounder must be a function.");
	const uint8_t totalDigits{
		static_cast<uint8_t>(minorUnitDigits(code) + currencyPrecision + extraFloatingPointPrecision)
	};
	const int32_t ratio{minorUnitPrecisionFactor(code)};
	feclearexcept(FE_ALL_EXCEPT);
	const UnitStorage major{
		static_cast<UnitStorage>(
			verifiedFloatingPoint((rounder(value, totalDigits) * ratio) / ratio)
		)
	};
	return major;
}

MinorUnit pecunia::currency::internal::fromFloatingPointMinor(
	const FloatingPoint& value,
	const Iso4217Codes& code,
	const uint8_t currencyPrecision,
	const FloatingPointRounderFunction& rounder
)
{
	assert(rounder != nullptr && "The floating point rounder must be a function.");
	const uint8_t totalDigits{
		static_cast<uint8_t>(minorUnitDigits(code) + currencyPrecision + extraFloatingPointPrecision)
	};
	const int32_t ratio{
		minorUnitPrecisionFactor(code)
			* static_cast<int32_t>(pow(10, extraFloatingPointPrecision))
	};
	feclearexcept(FE_ALL_EXCEPT);
	const UnitStorage adjustedValue{
		static_cast<UnitStorage>(
			verifiedFloatingPoint(rounder(value, totalDigits) * ratio)
		)
	};
	const auto minorAdjusted{abs(adjustedValue) % ratio};
	const auto extraRatio{pow(10.0, extraFloatingPointPrecision)};
	const auto removedExtra{minorAdjusted / static_cast<MinorUnit>(extraRatio)};
	return static_cast<MinorUnit>(removedExtra);
}

UnitStorage pecunia::currency::internal::normaliseAmount(const Money& from, const Iso4217Codes& to)
{
	assert(pecunia::converter != nullptr && "The currency conversion function is not set.");
	const auto& fromCode{from.code()};
	const FloatingPoint conversionFactor{pecunia::converter(fromCode, to)};

	if (isEqual(conversionFactor, 1.0, static_cast<uint8_t>(minorUnitDigits(fromCode) + precision)))
		return from.amount();
	feclearexcept(FE_ALL_EXCEPT);
	const auto convertedAmount{
		verifiedFloatingPoint(static_cast<FloatingPoint>(from.amount()) * conversionFactor)
	};
	return static_cast<UnitStorage>(convertedAmount);
}
