/*******************************************************************************
***  This file is part of Pecunia.                                           ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021                              ***
***  John Schneiderman <Licensing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Lesser General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Lesser General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Lesser General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#include "Verification.h"
using std::int32_t;
using std::uint8_t;
using pecunia::precision;
using pecunia::MinorUnit;
using pecunia::FloatingPoint;
using pecunia::UnitStorage;
using pecunia::internal::VerifiedValue;

#include <cmath>
using std::fabs;
using std::pow;
#include <cfenv>
using std::feclearexcept;
using std::fetestexcept;
#include <stdexcept>
using std::domain_error;
using std::overflow_error;
using std::underflow_error;


bool pecunia::internal::isEqual(
	const FloatingPoint& number,
	const FloatingPoint& equals,
	const uint8_t minorPrecision
)
{
	const FloatingPoint epsilon{1 / pow(10, minorPrecision + 1)};
	const FloatingPoint difference{fabs(number - equals)};
	return difference < epsilon;
}

FloatingPoint pecunia::internal::verifiedFloatingPoint(const FloatingPoint& value)
{
	int32_t exceptions{fetestexcept(FE_ALL_EXCEPT)};
	feclearexcept(FE_ALL_EXCEPT);

	if (exceptions & FE_DIVBYZERO)
		throw domain_error{"The floating-point operation resulted in a divided by zero."};

	if (exceptions & FE_UNDERFLOW)
		throw underflow_error{
			"The floating-point operation resulted in a value that is too small to fit within "
				"the FloatingPoint type."
		};

	if (exceptions & FE_OVERFLOW)
		throw overflow_error{
			"The floating-point operation resulted in a value that is too large to fit within "
				"the FloatingPoint type."
		};

	if (exceptions & FE_INVALID)
		throw domain_error{
			"The floating-point operation resulted in a value that is not valid."
		};
	return value;
}

VerifiedValue pecunia::internal::verifyAdditionFits(
	const UnitStorage& lhs,
	const UnitStorage& rhs,
	const UnitStorage& maximum,
	const UnitStorage& minimum
)
{
	const UnitStorage maximumDifference{static_cast<UnitStorage>(maximum - rhs)};
	if ((rhs > 0) && (lhs > maximumDifference))
		return VerifiedValue::Overflows;

	const UnitStorage minimumDifference{static_cast<UnitStorage>(minimum - rhs)};
	if ((rhs < 0) && (lhs < minimumDifference))
		return VerifiedValue::Overflows;
	return VerifiedValue::Fits;
}

VerifiedValue pecunia::internal::verifySubtractionFits(
	const UnitStorage& lhs,
	const UnitStorage& rhs,
	const UnitStorage& maximum,
	const UnitStorage& minimum
)
{
	const UnitStorage maximumDifference{static_cast<UnitStorage>(maximum + rhs)};
	if ((rhs < 0) && (lhs > maximumDifference))
		return VerifiedValue::Overflows;

	const UnitStorage minimumDifference{static_cast<UnitStorage>(minimum + rhs)};
	if ((rhs > 0) && (lhs < minimumDifference))
		return VerifiedValue::Overflows;
	return VerifiedValue::Fits;
}

VerifiedValue pecunia::internal::verifyMultiplicationFits(
	const UnitStorage& lhs,
	const FloatingPoint& rhs,
	const UnitStorage& maximum,
	const UnitStorage& minimum,
	const MinorUnit& minorMaximum,
	const uint8_t minorPrecision
)
{
	if ((lhs == 0) || isEqual(rhs, 0, minorPrecision))
		return VerifiedValue::Fits;

	if ((lhs == -1) || isEqual(rhs, -1, minorPrecision))
		return VerifiedValue::Fits;

	const auto fpLhs{static_cast<FloatingPoint>(lhs)};
	const auto fpMaximum{static_cast<FloatingPoint>(maximum)};
	if (fpLhs > fpMaximum / rhs)
		return VerifiedValue::Overflows;

	const auto fpMinimum{static_cast<FloatingPoint>(minimum)};
	if (fpLhs < fpMinimum / rhs)
		return VerifiedValue::Overflows;

	if (rhs < 1 && rhs > -1)
	{
		const FloatingPoint reciprocal{1 / rhs};
		return verifyDivisionFits(lhs, reciprocal, maximum, minimum, minorMaximum, minorPrecision);
	}
	return VerifiedValue::Fits;
}

VerifiedValue pecunia::internal::verifyDivisionFits(
	const UnitStorage& lhs,
	const FloatingPoint& rhs,
	const UnitStorage& maximum,
	const UnitStorage& minimum,
	const MinorUnit& minorMaximum,
	const uint8_t minorPrecision
)
{
	if (isEqual(rhs, 0, minorPrecision))
		return VerifiedValue::Undefined;

	if (rhs < 1 && rhs > -1)
	{
		const FloatingPoint reciprocal{1 / rhs};
		return verifyMultiplicationFits(
			lhs,
			reciprocal,
			maximum,
			minimum,
			minorMaximum,
			minorPrecision
		);
	}

	const auto fpLhs{static_cast<FloatingPoint>(lhs)};
	if ((lhs > 0) && (lhs <= minorMaximum) && (rhs > fpLhs))
		return VerifiedValue::Underflows;
	return VerifiedValue::Fits;
}

VerifiedValue pecunia::internal::verifyMultiplicationFits(
	const UnitStorage& lhs,
	const UnitStorage& rhs,
	const UnitStorage& maximum,
	const UnitStorage& minimum
)
{
	if ((lhs == 0) || rhs == 0)
		return VerifiedValue::Fits;

	if ((lhs == -1) || rhs == -1)
		return VerifiedValue::Fits;

	if (lhs > maximum / rhs)
		return VerifiedValue::Overflows;

	if (lhs < minimum / rhs)
		return VerifiedValue::Overflows;
	return VerifiedValue::Fits;
}

VerifiedValue pecunia::internal::verifyDivisionFits(
	const UnitStorage& lhs,
	const UnitStorage& rhs,
	const MinorUnit& minorMaximum
)
{
	if (rhs == 0)
		return VerifiedValue::Undefined;

	if (lhs > 0 && lhs <= minorMaximum && rhs > lhs)
		return VerifiedValue::Underflows;
	return VerifiedValue::Fits;
}
